import os
from jinja2 import Environment

from cpbox.tool import file
from cpbox.tool import utils

def render_to_str(template_file, data):
    if not os.path.isfile(template_file):
        raise Exception('template file not found: ' +  template_file)
    result = ''
    with open(template_file, 'r') as f:
        patten = f.read()
        result = render_str(patten, data)
    return result

def render_to_file(template_file, data, target_file):
    if not os.path.isfile(template_file):
        raise Exception('template file not found: ' +  template_file)
    file.ensure_dir_for_file(target_file)
    with open(target_file, 'w') as fw:
        fw.write(render_to_str(template_file, data))

def render(template_file, config_file=None, config_section=None, extra_kvs=None, output_file=None, output_dir=None, **kwargs):
    data = utils.load_yaml_config(config_file, config_section, extra_kvs)
    if not output_file and not output_dir:
        raise Exception('should specify output_file or output_dir')

    if output_file is not None:
        output_file = output_file
    else:
        output_file = output_dir.rstrip('/') + '/' + os.path.basename(template_file)
    return render_to_file(template_file, data, output_file)

def render_str(patten, data):
    env = Environment(keep_trailing_newline=True)
    patten = patten.decode('utf-8')
    s = env.from_string(patten)
    ret = s.render(data)
    return ret



# coding: utf-8
import sys
is_py2 = sys.version[0] == '2'
if is_py2:
    from Queue import Queue
else:
    from queue import Queue
from threading import Timer, Lock, Thread

import copy
from cpbox.tool import array
from cpbox.tool import utils
import logging
import time
logger = logging.getLogger()

_DEBUG_CONCURRENT = False

def waiting(get_fn, check_should_wait_fn, on_waiting_fn, interval=5, wait_before_start=0):
    if wait_before_start:
        time.sleep(wait_before_start)
    while True:
        ret = get_fn()
        should_wait = check_should_wait_fn(ret)
        if should_wait:
            if callable(on_waiting_fn):
                on_waiting_fn(ret)
            time.sleep(interval)
        else:
            return ret

class Worker(object):
    def __init__(self, num_worker_threads = 2):
        self.queue = Queue()
        for i in range(num_worker_threads):
             t = Thread(target=self.do_work)
             t.daemon = True
             t.start()

    def do_work(self):
        if _DEBUG_CONCURRENT:
            logger.debug('begin_do_work')
        while True:
            callback, data = self.queue.get()
            callback(data)
            self.queue.task_done()

    def submit(self, callback, data):
        self.queue.put((callback, data))

    def wait(self):
        self.queue.join()

class RepeatedTimer(object):
    def __init__(self, interval, function, *args, **kwargs):
        self._timer = None
        self.interval = interval
        self.function = function
        self.args = args
        self.kwargs = kwargs
        self.is_running = False
        self.start()

    def _run(self):
        self.is_running = False
        self.start()
        self.function(*self.args, **self.kwargs)

    def start(self):
        if not self.is_running:
            self._timer = Timer(self.interval, self._run)
            self._timer.daemon = True
            self._timer.start()
            self.is_running = True

    def stop(self):
        self._timer.cancel()
        self.is_running = False

class DelayOp:

    def __init__(self, callback, timeout = 3, threshold = 10):
        self.callback = callback
        self.timeout = timeout
        self.threshold = threshold
        self.data_list = []
        self.lock = Lock()
        self.skip_timer_check_for_one_round = False
        self.timer = RepeatedTimer(self.timeout, self._do_timer_check)
        self.worker = Worker(4)

    def _do_timer_check(self):
        if _DEBUG_CONCURRENT:
            logger.debug('_do_timer_check')
        if self.skip_timer_check_for_one_round:
            self.skip_timer_check_for_one_round = False
            if _DEBUG_CONCURRENT:
                logger.debug('_do_timer_check skipped')
            return
        self.lock.acquire()
        self._do_check(True)
        self.lock.release()

    def _do_check(self, skip_threshold_check_for_timer = False):
        if _DEBUG_CONCURRENT:
            logger.debug('_do_check')
        size = len(self.data_list)
        reach_threshold = size >= self.threshold
        if reach_threshold:
            self.skip_timer_check_for_one_round = True
        if reach_threshold or (size > 0 and skip_threshold_check_for_timer):
            list = copy.deepcopy(self.data_list)
            self.data_list = []
            for sub_list in array.chunks(list, self.threshold):
                self.worker.submit(self.callback, sub_list)

    def add_delay_item(self, item):
        if _DEBUG_CONCURRENT:
            logger.debug('add_delay_item: %s', item)
        self.lock.acquire()
        self.data_list.append(item)
        self._do_check(False)
        self.lock.release()



import unittest
from cpbox.tool import redistool

class TestToolredis(unittest.TestCase):

    def setUp(self):
        url = 'redis://:%s@%s:%s' % ('cp', '172.16.1.102', '10300')
        redistool.init_proxy(url)

    def test_get_kv(self):
        k = 'test'
        v = {'test': 'test'}
        value = redistool.get(k)
        self.assertEqual(value, False)

        redistool.set(k, v)
        value = redistool.get(k)
        self.assertEqual(value, v)
        redistool.delete(k)
        value = redistool.get(k)
        self.assertEqual(value, False)

    def test_get_list(self):
        key_list = []
        key_list.append('test1')
        key_list.append('test3')
        key_list.append('test2')
        redistool.delete_keys(key_list)
        value_list = redistool.get_list(key_list)
        self.assertTrue(not value_list)

        kv_list = {}
        kv_list['test1'] = {'test1': 'test1'}
        kv_list['test3'] = {'test3': 'test3'}
        kv_list['test2'] = {'test2': 'test2'}
        redistool.set_list(kv_list)

        want_value = {'test3': {'test3': 'test3'}, 'test1': {'test1': 'test1'}, 'test2': {'test2': 'test2'}}
        value_list = redistool.get_list(key_list)
        self.assertEqual(want_value, value_list)

        redistool.delete_keys(key_list)
        value_list = redistool.get_list(key_list)
        self.assertTrue(not value_list)

if __name__ == '__main__':
    unittest.main()



import os
import fcntl
import shutil
from os import path

_STR_MAP = {
        ' ':'.',
        '/': '-',
        '--': '-',
        }

def listdir(dir, relative_path=False):
    for root, directories, filenames in os.walk(dir):
        for filename in filenames:
            filepath = path.join(root, filename)
            if relative_path:
                filepath = filepath.replace(dir, '')
            yield filepath

def ensure_dir(dir, mode=0o777):
    if not os.path.exists(dir):
        oldmask = os.umask(000)
        os.makedirs(dir, mode)
        os.umask(oldmask)

def ensure_dir_for_file(file, mode=0o777):
    dir = os.path.dirname(file)
    ensure_dir(dir, mode)

def compute_lock_filepath(sys_argv):
    filepath = ' '.join(sys_argv)
    for old, new in _STR_MAP.items():
        filepath = filepath.replace(old, new)
    return filepath + '.lock'

def obtain_lock(filepath):
    ensure_dir_for_file(filepath)
    pidfile = open(filepath, 'a+')
    try:
        fcntl.flock(pidfile.fileno(), fcntl.LOCK_EX | fcntl.LOCK_NB)
    except IOError:
        pidfile = None
        return None
    pidfile.seek(0)
    pidfile.truncate()
    pidfile.write(str(os.getpid()))
    pidfile.flush()
    pidfile.seek(0)
    return pidfile

def rmtree(folder_path):
    for file_object in os.listdir(folder_path):
        file_object_path = os.path.join(folder_path, file_object)
        if os.path.isfile(file_object_path) or os.path.islink(file_object_path):
            os.unlink(file_object_path)
        else:
            shutil.rmtree(file_object_path)

def copytree(src, dst, symlinks=False, ignore=None):
    for file_object in os.listdir(src):
        src_file_object_path = os.path.join(src, file_object)
        dst_file_object_path = os.path.join(dst, file_object)
        if os.path.isfile(src_file_object_path):
            shutil.copy(src_file_object_path, dst_file_object_path)
        elif os.path.islink(src_file_object_path):
            linkto = os.readlink(src_file_object_path)
            os.symlink(linkto, dst_file_object_path)
        else:
            shutil.copytree(src_file_object_path, dst_file_object_path, symlinks, ignore)

def free_space(path):
    if not os.path.isdir(path):
        return 0
    s = os.statvfs(path)
    free_space = (s.f_bavail * s.f_frsize)
    return free_space



import socket
import platform
import os

from os import path
from cpbox.tool import template
import docker

def get_container(name):
    client = docker.from_env()
    container = None
    try:
        container = client.containers.get(name)
    except docker.errors.NotFound:
        pass
    return container

def remove_container(name, **kwargs):
    container = get_container(name)
    if container:
        container.remove(**kwargs)

def stop_container(name, **kwargs):
    container = get_container(name)
    if container:
        container.stop(**kwargs)

def get_os_name():
    os_name = platform.system()
    map = {
            'Linux': 'linux',
            'Darwin': 'mac',
            'CYGWIN': 'cygwin',
            'MINGW': 'MinGw',
            }
    for name, sys in map.iteritems():
        if os_name.startswith(name):
            return sys
    return 'unknown'

def base_docker_args(env=None, container_name=None, volumes=None, ports=None, current_user=False, envs=None, working_dir=None):
    hostname = socket.gethostname()
    args = '--name ' + container_name
    args += ' -h %s.docker.%s' % (container_name, hostname)
    args += ' -e PUPPY_HOSTNAME=' + hostname
    args += ' -e PUPPY_CONTAINER_NAME=' + container_name
    for key, value in os.environ.items():
        if key.startswith('PUPPY_'):
            args += ' -e %s="%s"' % (key, value)
    if env:
        args += ' -e PUPPY_ENV=' + env

    if working_dir and os.path.isdir(working_dir):
        args += ' -w ' + working_dir

    if current_user:
        args += ' -u %s:%s' % (os.getuid(), os.getgid())

    if volumes:
        args += ' ' + ' '.join(('-v ' + from_ + ':' + to_ for from_, to_ in volumes.iteritems()))
    if ports:
        if isinstance(ports, dict):
            args += ' ' + ' '.join(('-p ' + str(from_) + ':' + str(to_) for from_, to_ in ports.iteritems()))
        else:
            args += ' ' + ' '.join(('-p ' + str(item) for item in ports))
    if envs:
        args += ' ' + ' '.join(('-e ' + item for item in envs))

    os_name = get_os_name()
    if os_name != 'mac':
        args += ' -v /etc/localtime:/etc/localtime'
    if path.isfile('/etc/devops'):
        args += ' -v /etc/devops/:/etc/devops/:ro'

    return args



def chunks(l, n):
    """Yield successive n-sized chunks from l."""
    for i in range(0, len(l), n):
        yield l[i:i + n]

def merge(src, update):
    """
    >>> a = { 'first' : { 'all_rows' : { 'pass' : 'dog', 'number' : '1' } } }
    >>> b = { 'first' : { 'all_rows' : { 'fail' : 'cat', 'number' : '5' } } }
    >>> merge(a, b) == { 'first' : { 'all_rows' : { 'pass' : 'dog', 'fail' : 'cat', 'number' : '5' } } }
    True
    """
    for key, value in update.items():
        if isinstance(value, dict):
            # get node or create one
            node = src.setdefault(key, {})
            merge(node, value)
        else:
            src[key] = value

    return src



import unittest
from cpbox.tool import concurrent
from threading import Thread
import time
import logging
log = logging.getLogger()

class TestConcurrent(unittest.TestCase):

    def setUp(self):
        self.num = 0
        self.callback_items = []

    def tearDown(self):
        pass

    def test_deplay_op(self):
        num_max = 25
        put_item_interval = 0.01
        timeout = 0.5
        threshold = 10

        self.callback_items = []
        def callback(items):
            if concurrent.debug_concurrent:
                log.debug('callback: %s', len(items))
            self.callback_items.extend(items)

        def wait():
            while len(self.callback_items) != num_max:
                time.sleep(put_item_interval * 3)

        def add_element():
            # nonlocal num
            while self.num < num_max:
                time.sleep(put_item_interval)
                self.num = self.num + 1
                delay_op.add_delay_item(self.num)

        delay_op = concurrent.DelayOp(callback, timeout, threshold=threshold)
        thread_add_element = Thread(target=add_element)
        thread_add_element.daemon = True
        thread_add_element.start()

        thread_wait = Thread(target=wait)
        thread_wait.daemon = False
        thread_wait.start()

if __name__ == '__main__':
    unittest.main()



from setuptools import setup, find_packages

setup(
    name = 'cpbox',
    version = '0.0.26',
    keywords = ('cpbox'),
    description = 'cp tool box',
    license = '',
    install_requires = [
        'six',
        'pyyaml',
        'Jinja2',
        'netaddr',
        'requests',
        'pytz>=3.6',
        'urllib3[secure]',
        'cryptography>=2.2.1',
        'docker',
        'redis'
        ],

    scripts = [],

    author = 'http://www.liaohuqiu.net',
    author_email = 'liaohuqiu@gmail.com',
    url = '',

    packages = find_packages(),
    platforms = 'any',
)



from datetime import datetime
from pytz import reference
from tzlocal import get_localzone

from cpbox.app.appconfig import appconfig
from cpbox.tool import concurrent

import time
import os
import threading
import json
import logging
import socket

try:
    from threading import get_ident
except ImportError:
    from thread import get_ident

event_logger = logging.getLogger('event-log')
local_timezone = get_localzone()
local_ip = socket.gethostbyname(socket.gethostname())

log_worker = concurrent.Worker(2)
# ${time_iso_8601} ${client_ip} ${pid}.${rnd/thread_id} {$app_name} ${event_key} ${payload_json_encoded}
def add_event_log(event_key, payload):
    payload['env'] = appconfig.get_env()
    time = local_timezone.localize(datetime.now())
    time = time.strftime('%Y-%m-%dT%H:%M:%S%z')
    msg = '%s %s %s.%s %s %s %s' % (time, local_ip, os.getpid(), get_ident(), appconfig.get_app_name(), event_key, json.dumps(payload))
    log_worker.submit(event_logger.info, msg)

def log_func_call(func, *args, **kwargs):
    def timed(*args, **kw):
        start = time.time() * 1000
        result = func(*args, **kw)
        payload = {}
        payload['name'] = func.__name__
        payload['rt'] = time.time() * 1000 - start
        add_event_log('func-call', payload)
        return result
    return timed
